/*
 * Copyright (c) 2022 VMware, Inc. All Rights Reserved.
 */
const https = require('https');
const {gatewayConfig, smDeployment} = require('./lib/config');
const tracing = require('./lib/tracing.js');
const express = require('express');
const compression = require('compression');
const {createYoga} = require('graphql-yoga');
const {useHive} = require('@graphql-hive/client');
const {createFetch} = require('@whatwg-node/fetch');
const {useExtendedValidation} = require('@envelop/extended-validation');
const expressWinston = require('express-winston');

const {stitchSchemas} = require('@graphql-tools/stitch');
const {stitchingDirectives} = require('@graphql-tools/stitching-directives');
const {stitchingDirectivesTransformer} = stitchingDirectives();
const {printSchema} = require('graphql');
const {printSchemaWithDirectives} = require("@graphql-tools/utils");
const {altairExpress} = require('altair-express-middleware');
const fs = require('fs');
const {createHandler} = require('graphql-sse/lib/use/express');

const SchemaLoader = require('./lib/schema_loader');
const makeRemoteExecutor = require('./lib/make_remote_executor');

const {log, requestLogger} = require("./lib/logger");
const {setRequestExecutionContext, reqCtxt} = require("./lib/request_context");
const {setAuthContextForRequest} = require("./lib/auth");
const {createPrivateGraphQLSchema, createFilteredPublicGraphQLSchema} = require('./lib/schema_filter_service');
const handleHealthCheck = require('./lib/health_service');
const checkQueryComplexity = require('./lib/queryComplexityRule');
const {pruneSchema} = require("@graphql-tools/utils");

const schemaAdditions = fs.readFileSync("./schemas/local/schemaAdditions.graphqls", "utf-8");
let loadSchemaFromFS = process.env.LOAD_SCHEMA_FROM_FS || false;
expressWinston.requestWhitelist.push('body');

const defaultQuery = fs.readFileSync(smDeployment ? "sm_default_query.graphql" : "default_query.graphql", 'utf8');
const defaultAltairQuery = fs.readFileSync(smDeployment ? "sm_default_query_altair.graphql" : "default_query_altair.graphql", 'utf8');
const objectHash = require('object-hash');
const axios = require("axios");
const url = require("url");

//Socket Timeout constant to 20 min
const SOCKET_TIMEOUT = 20 * 60 * 1000;

const UPLOAD_MAX_FILE_SIZE = process.env.UPLOAD_MAX_FILE_SIZE || 10 * 1024 * 1024;
const UPLOAD_MAX_FILES = process.env.UPLOAD_MAX_FILES || 5;

// load gateway config
const config = gatewayConfig();

const loader = new SchemaLoader({
    loadSchemaFromFS: loadSchemaFromFS,
    endpoints: config.endpoints,
    stitchSchemas: (loadedEndpoints) => {
        const subschemas = [];
        for (const [endpoint, schema] of loadedEndpoints.entries()) {
            if (endpoint.customMerge != null) {
                // TODO ENS-6296 this should be data-driven on a per-microservice basis to a library of merge functions
                merger = getMergeConfig(endpoint.customMerge);
            } else {
                merger = null;
            }
            subschemas.push({
                executor: makeRemoteExecutor(endpoint.url, endpoint.subscriptionUrl, {timeout: config.remoteTimeoutms}),
                schema: schema,
                merge: merger,
                batch: true,
            });
        }
        ;

        return stitchSchemas({
            subschemaConfigTransforms: [stitchingDirectivesTransformer],
            subschemas: subschemas,
            // if required, work around any schema error(s)
            typeMergingOptions: {
                validationLevel: 'error'
            },
            // Expose SDL directly at this level - TODO make String! but must also modify ensemble-commons definition
            // Public sdl query does not expose directives (does not use printSchemaWithDirectives) since directives can contain internal information
            // If we need to expose information contained in directives to end users (e.g. auth permissions) we should add specific queries
            typeDefs: schemaAdditions,
            resolvers: {
                Query: {
                    'sdl': (_, args, context) => printSchema(getSchemaBasedOnEndPoint(context.req)),
                    '_sdl': () => printSchemaWithDirectives(pruneSchema(loader.schema)),
                    '_subSchemasSdl': () => getSubSchemas(loader.loadedEndpoints),
                    '_queryHealthStatus': handleHealthCheck
                }
            }
        });

        function getSubSchemas(loadedEndpoints) {
            const result = [];
            for (const [endpoint, subSchema] of loadedEndpoints.entries()) {
                result.push({
                    name: endpoint.name,
                    schema: printSchema(subSchema)
                });
            }
            return result;
        }
    }
});

function getSchemaBasedOnEndPoint(request) {
    let originalUrl = request?.originalUrl || request?.raw?.originalUrl;
    let isPrivateEndpoint = originalUrl.startsWith(`/${config.hubEndpointName}-private`)
        || originalUrl.startsWith(`/${config.ariaEndpointName}-private`)
        || originalUrl.startsWith(`/${config.ensembleEndpointName}-private`);
    let graphQLSchema = isPrivateEndpoint ? createPrivateGraphQLSchema(loader.schema) : createFilteredPublicGraphQLSchema(loader.schema);
    return graphQLSchema;
}

function extractUpstreamRequiredHeaders(req) {
    let headers = {};
    //the subscriptions handler wraps the request, express req has req.get whereas graphql-sse wrapped req has req.headers.get function.
    if (typeof req.get !== 'function') {
        req.get = req.headers.get;
    }
    if (req.get('authorization')) {
        headers['authorization'] = req.get('authorization')
    }
    if (req.get('accept-language')) {
        headers['accept-language'] = req.get('accept-language')
    }
    if (req.get('x-graphql-client-name')) {
        headers['x-graphql-client-name'] = req.get('x-graphql-client-name')
    }
    //header for auth context
    if (req.get('x-authorization-context')) {
        headers['x-authorization-context'] = req.get('x-authorization-context')
    }
    return headers;
}

function createYogaServer(endpoint) {
    return createYoga({
        fetchAPI: createFetch({
            // We prefer `node-fetch` over `undici` and current unstable Node's implementation
            useNodeFetch: true,
            formDataLimits: {
                // Maximum allowed file size (in bytes)
                fileSize: UPLOAD_MAX_FILE_SIZE,
                // Maximum allowed number of files
                files: UPLOAD_MAX_FILES,
            }
        }),
        schema: async ({req}) => {
            return getSchemaBasedOnEndPoint(req);
        },
        context: async ({params, req}) => {
            let context = {};
            context.headers = extractUpstreamRequiredHeaders(req);
            return context;
        },
        graphiql: {
            defaultQuery: defaultQuery,
            subscriptionsProtocol: "SSE",
            title: `VMware Tanzu Hub - GraphQL interface™`
        },
        landingPage: false,
        graphqlEndpoint: endpoint,
        multipart: true,
        maskedErrors: false, //We use got for our remote executors, and it cannot handle GraphqlError objects. Yoga needs GraphqlError objects
        //to send back custom messages, if not it just sends out an Unexpected Error message. We already have error handling in place to handle
        //errors in the remote executor. This will ensure Yoga doesn't meddle with the errors and our error handling comes into play.
        plugins: [
            useExtendedValidation({
                rules: [
                    checkQueryComplexity
                ]
            }),
            useHive({
                enabled: config.hive.metrics.enabled,
                token: config.hive.metrics.token,
                usage: {
                    sampleRate: config.hive.metrics.sampleRate,
                    clientInfo(context) {
                        //client name will be x-graphql-client-name head value or user name or ip
                        var name = context.req.headers['x-graphql-client-name']
                        if (!name) {
                            var name = reqCtxt.get("user") || context.req.headers['x-forwarded-for'] || context.req.ip;
                            if (name && config.hive.metrics.maskUserName) {
                                // generationg the hash code from the user name to mask the user name
                                name = objectHash(name);
                            }
                        }

                        const version = reqCtxt.get("orgId") || context.headers['x-graphql-client-version'] || "unknown";
                        if (name) {
                            return {name, version}
                        }
                        return null
                    }
                }
            })
        ]
    });
}


loader.reload().then(() => {
    const app = express();
    app.use(setRequestExecutionContext);
    app.use(setAuthContextForRequest);
    app.use(express.json({limit: "10mb"})); //Heimdall supports a max of 10MB payload.
    app.disable("etag");
    app.disable("x-powered-by");
    app.use(middleWareErrorHandler)
    app.use(requestLogger());
    app.use(compression())
    app.use(express.urlencoded())

    let subscriptionHandler = createHandler({
        schema: async (req, executionArgsWithoutSchema) => {
            return getSchemaBasedOnEndPoint(req);
        },
        context: async (req, args) => {
            let subscriptionCtxt = {};
            subscriptionCtxt.headers = extractUpstreamRequiredHeaders(req);
            reqCtxt.set('subscriptionCtxt', subscriptionCtxt);
            return subscriptionCtxt;
        }
    })

    const endpoints = [
        `/${config.hubEndpointName}`,
        `/${config.ariaEndpointName}`,
        `/${config.ensembleEndpointName}`
    ]

    for (const endpoint of endpoints) {
        // Create GraphQL endpoints
        // Need to do it this way see  https://github.com/dotansimha/graphql-yoga/issues/1901
        // Without this altair works but not the yoga bundled graphiql interface.
        app.use(endpoint + '/graphql', createYogaServer(endpoint + '/graphql'));
        app.use(endpoint + '-private/graphql', createYogaServer(endpoint + '-private/graphql'));

        app.use(endpoint + '/subscriptions', subscriptionHandler);

        // static HTML docs
        app.use(endpoint + "/docs", express.static('docs'));

        // Altair GraphQL client
        app.use(endpoint + "/altair", altairExpress({
                endpointURL: endpoint + "/graphql",
                subscriptionsEndpoint: endpoint + "/subscriptions",
                initialQuery: defaultAltairQuery,
            })
        );

        // redirect base endpoint URL to docs
        app.get(endpoint, function (req, res) {
            res.redirect(endpoint + "/docs");
        });

        // add liveness and readiness probes
        app.use([
                endpoint + "/readiness",
                endpoint + "/liveness"
            ], function (req, res) {
                res.send("Health check passed");
            }
        );
    }
    if (config.oauth2) {
        app.get("/csp/gateway/discovery", (req, res) => {
            // Add pinniped related configuration
            let queryParams = Object.assign(req.query);
            queryParams.pinniped_idp_name = config.oauth2.idpName
            res.redirect(url.format({
                pathname: config.oauth2.authenticationUrl,
                query: queryParams,
            }));
        });

        app.post('/csp/gateway/am/api/auth/revoke', async (req, res) => {
            //TODO: Enable this revoke logic once we have proper revoke logic.
            // const tokenUrl = "https://localhost:48443/any/path/oauth2/revoke";
            // res.set('Access-Control-Allow-Origin', '*');
            // const headersAndBody = prepareHeadersAndRequestBody(req.body);
            // const body = await sendPostRequestAndRespond(tokenUrl, headersAndBody.formBody, headersAndBody.headers).then((access_token_response) => {
            //     res.send(access_token_response);
            // }).catch((err) => {
            //     res.send(err);
            // })
            res.send({"message": "revoked refresh token"});

        });

        app.post('/csp/gateway/am/api/auth/logout', async (req, res) => {
            const tokenUrl = config.oauth2.logoutUrl;
            res.set('Access-Control-Allow-Origin', '*');
            const headersAndBody = prepareHeadersAndRequestBody(req.body);
            const body = await sendPostRequestAndRespond(tokenUrl, headersAndBody.formBody, headersAndBody.headers).then((access_token_response) => {
                res.send(access_token_response);
            }).catch((err) => {
                res.send(err);
            })
        });

        app.post('/csp/gateway/am/api/auth/token', async (req, res) => {
            res.set('Access-Control-Allow-Origin', '*');
            const headersAndBody = prepareHeadersAndRequestBody(req.body);

            const tokenUrl = config.oauth2.authorizationUrl;
            const access_token_response = await sendPostRequestAndRespond(tokenUrl, headersAndBody.formBody, headersAndBody.headers);
            // interchanging id_token and access_token as csp oauth2 library expects
            // access_token in JWT format and it uses access_token for authentication.
            // Pinniped gives us access_token as a hashed string and id_token as JWT
            // To ensure no changes in CSP library we are interchanging the tokens here.
            try {
                const access_token = access_token_response.access_token;
                access_token_response.access_token = access_token_response.id_token;
                access_token_response.id_token = access_token;
                res.send(access_token_response);
            } catch (err) {
                res.send(err);
            }
        });
    }


    let server = app.listen(4000, () => log.info(`gateway running http://localhost:4000/${config.hubEndpointName}/graphql`));
    server.setTimeout(SOCKET_TIMEOUT);
    loader.autoRefresh();
});


async function sendPostRequestAndRespond(url, requestData, headers) {
    try {
        console.log("Sending Request", url, requestData, headers);
        const response = await axios.post(url, requestData, {headers: headers});
        return response.data;
    } catch (error) {
        console.log("Post request to Url " + url + " failed with error: ", error);
        throw error;
    }

}

function prepareHeadersAndRequestBody(req_body) {
    let formBody = [];
    Object.entries(req_body).forEach(([propertyKey, propertyValue]) => {
        formBody.push(encodeURIComponent(propertyKey) + "=" + encodeURIComponent(propertyValue));
    })
    formBody = formBody.join("&");
    const auth = 'Basic ' + new Buffer(config.oauth2.clientName + ':' + config.oauth2.clientSecret).toString('base64');
    const headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': auth
    }

    return {headers: headers, formBody: formBody};
}

function middleWareErrorHandler(error, req, res, next) {
    if (error) {
        log.error(`Error while processing request:  ${error.stack}`);
        // some errors have right http status codes whereas some will be a generic error/exception.
        // For generic error we send back a 500 status response to avoid leaking impl details.
        const {status: statusCode = 500, message = "Internal Server Error"} = error;
        res.status(statusCode).send({"message": message});
    } else {
        next();
    }
}

process
    .on('unhandledRejection', (reason, p) => {
        console.error(reason, 'Unhandled Rejection at Promise', p);
    })
    .on('uncaughtException', err => {
        console.error(err, 'Uncaught Exception thrown');
    });

// the below really belongs in its own library file of custom merge functions
function getMergeConfig(name) {
    if (name == "userMerge") {
        return {
            User: {
                entryPoints: [
                    {
                        selectionSet: '{ userId }',
                        fieldName: '_queryUsersByKeys',
                        key: ({userId}) => ({userId}),
                        argsFromKeys: (keys) => ({keys})
                    },
                    {
                        selectionSet: '{ userAccount }',
                        fieldName: '_queryUsersByKeys',
                        key: ({userAccount}) => ({userAccount}),
                        argsFromKeys: (keys) => ({keys})
                    }
                ]
            }
        }
    } else if (name == "entityMerge") {
        return {
            Entity: {
                selectionSet: '{ entityId }',
                fieldName: '_queryEntitiesByEntityIds',
                key: ({entityId}) => (entityId),
                argsFromKeys: (entityIds) => ({entityIds})
            }
        }
    }
    return null;
}
